/****************************************************************************
**
** Copyright (C) 2013 Jolla Ltd.
** Contact: Petri M. Gerdt <petri.gerdt@jollamobile.com>
**
****************************************************************************/

import QtQuick 2.1
import org.nemomobile.lipstick 0.1
import Sailfish.Silica 1.0
import Sailfish.Silica.private 1.0
import Sailfish.Lipstick 1.0
import org.nemomobile.dbus 1.0

SilicaGridView {
    id: gridview

    property Item reorderItem
    property Item movingItem
    property int columns: 5
    property int initialCellWidth: (desktop.width - Theme.paddingLarge * 2) / columns
    property real firstRowOpacity: Math.min(Math.max((dashboard.contentY - dashboard.lockscreenAndStatusAreaHeight)/cellHeight, 0.0), 1.0)
    property bool launcherEditMode: desktop.removeApplicationEnabled
    property var launcherModel: model
    property bool rootFolder
    property int folderIndex: -1
    property QtObject folderComponent
    property Item folderItem
    property Item openedChildFolder

    width: cellWidth * columns
    anchors.horizontalCenter: parent.horizontalCenter
    // Increase cellWidth so that icon vertical edges are Theme.paddingLarge away from display edges
    cellWidth: Math.round(initialCellWidth + (initialCellWidth - Theme.iconSizeLauncher * 0.8) / (columns - 1))
    cellHeight: Math.round(Screen.height / 6)

    function categoryQsTrIds() {
        //% "AudioVideo"
        QT_TRID_NOOP("lipstick-jolla-home-folder_audiovideo")
        //% "Audio"
        QT_TRID_NOOP("lipstick-jolla-home-folder_audio")
        //% "Video"
        QT_TRID_NOOP("lipstick-jolla-home-folder_video")
        //% "Development"
        QT_TRID_NOOP("lipstick-jolla-home-folder_development")
        //% "Education"
        QT_TRID_NOOP("lipstick-jolla-home-folder_education")
        //% "Game"
        QT_TRID_NOOP("lipstick-jolla-home-folder_game")
        //% "Graphics"
        QT_TRID_NOOP("lipstick-jolla-home-folder_graphics")
        //% "Network"
        QT_TRID_NOOP("lipstick-jolla-home-folder_network")
        //% "Office"
        QT_TRID_NOOP("lipstick-jolla-home-folder_office")
        //% "Science"
        QT_TRID_NOOP("lipstick-jolla-home-folder_science")
        //% "Settings"
        QT_TRID_NOOP("lipstick-jolla-home-folder_settings")
        //% "System"
        QT_TRID_NOOP("lipstick-jolla-home-folder_system")
        //% "Utility"
        QT_TRID_NOOP("lipstick-jolla-home-folder_utility")
    }

    function showFolder(folder) {
        if (openedChildFolder) {
            // should never happen
            openedChildFolder.close()
        }

        if (!folderComponent) {
            folderComponent = Qt.createComponent("LauncherFolder.qml")
            if (folderComponent.status == Component.Error) {
                console.log("Error opening folder", folderComponent.errorString())
                return
            }
        }
        openedChildFolder = folderComponent.createObject(desktop, { 'model': folder })
    }

    function setEditMode(enabled) {
        if (movingItem && !enabled) {
            movingItem.cancelReordering()
        }
        desktop.removeApplicationEnabled = enabled
    }

    Image {
        id: newFolderIcon
        property bool active
        property Item target
        property bool show: active && rootFolder && folderIndex >= 0
        property string iconId: "icon-launcher-folder-01"
        source: "image://theme/" + iconId
        opacity: show ? 1.0 : 0.0
        Behavior on opacity { FadeAnimation {} }
        anchors.centerIn: target ? target : gridview
        anchors.verticalCenterOffset: target ? -target.iconOffset/2 : 0
        scale: 1.3
        parent: gridview.contentItem
        z: -3

        onShowChanged: if (show) {
            var index = Math.floor(Math.min(Math.random() * 16, 15))
            iconId = "icon-launcher-folder-" + (index >= 9 ? (index + 1) : "0" + (index + 1))
        }
    }

    Timer {
        id: startReorderTimer
        interval: 150
        onTriggered: movingItem.startReordering()
    }

    Timer {
        id: reorderTimer
        interval: folderItem && folderItem.itemType == LauncherModel.Folder ? 10 : 150
        onTriggered: movingItem.doReordering()
    }

    Connections {
        target: Lipstick.compositor
        onDisplayOff: setEditMode(false)
        onHomeActiveChanged: if (!Lipstick.compositor.homeActive) setEditMode(false)
    }

    delegate: Item {
        id: wrapper
        property int modelIndex: index
        property bool reordering
        property int newIndex: -1
        property int newFolderIndex: -1
        property real oldY
        property bool down: !reordering && !launcherEditMode && launcherItem.pressed && launcherItem.containsMouse
        property int itemType: model.object.type
        property alias iconOffset: launcherIcon.y
        property int folderItemCount: itemType == LauncherModel.Folder && model.object ? model.object.itemCount : 0
        property bool launcherEditMode: gridview.launcherEditMode
        property Item uninstallButton
        property bool isUpdating: model.object.isUpdating
        property Item updatingItem
        property bool firstMove: true
        property var previousMovePosition
        property real pressX
        property real pressY
        property bool dragged

        onLauncherEditModeChanged: {
            if (launcherEditMode && !uninstallButton) {
                uninstallButton = uninstallButtonComponent.createObject(launcherItem)
            }
        }

        onIsUpdatingChanged: {
            if (isUpdating && !updatingItem) {
                updatingItem = updatingComponent.createObject(launcherItem)
            }
        }

        // Make sure the X and Y coordinates change when the gridview positions the item
        x: -2000
        y: -2000
        width: cellWidth
        height: cellHeight
        visible: false // I'm a placeholder. launcherItem does all the painting

        onXChanged: moveTimer.running = true
        onYChanged: moveTimer.running = true

        Connections {
            target: model.object
            ignoreUnknownSignals: true
            onItemRemoved: {
                var object = model.object
                if (object.itemCount === 1) {
                    // One item remains in folder. Replace folder with item.
                    var parentFolderIndex = object.parentFolder.indexOf(object)
                    object.parentFolder.moveToFolder(object.get(0), object.parentFolder, parentFolderIndex)
                    object.destroyFolder()
                }
            }
        }

        Timer {
            id: moveTimer
            interval: 1
            onTriggered: {
                if (firstMove) {
                    // Don't animate when created
                    launcherItem.x = wrapper.x
                    launcherItem.y = wrapper.y
                    firstMove = false
                    oldY = y
                } else if (!reordering) {
                    if (y != oldY) {
                        slideMoveAnim.stop()
                        fadeMoveAnim.start()
                        oldY = y
                    } else if (!fadeMoveAnim.running) {
                        slideMoveAnim.restart()
                    }
                }
            }
        }

        MouseArea {
            id: launcherItem

            // TODO: Use LauncherGridItem instead.

            function startReordering() {
                if (launcherEditMode && !dragged) {
                    reparent(desktop)
                    reorderItem = launcherItem
                    drag.target = launcherItem
                    z = 1000
                    reordering = true
                    dragged = true
                }
            }

            function doReordering() {
                if (newFolderIndex >= 0 && newFolderIndex !== index) {
                    if (folderItem.itemType == LauncherModel.Application) {
                        newFolderIcon.target = folderItem
                        newFolderIcon.active = true
                    } else {
                        newFolderIcon.active = false
                    }
                    gridview.folderIndex = newFolderIndex
                } else if (newIndex != -1 && newIndex !== index) {
                    launcherModel.move(index, newIndex)
                }

                newIndex = -1
            }

            function cancelReordering() {
                if (startReorderTimer.running) {
                    startReorderTimer.stop()
                }
                if (reordering) {
                    reordering = false
                    reorderTimer.stop()
                    if (launcher.launcherModifierObject)
                        launcher.launcherModifierObject.stopScrolling()
                    drag.target = null
                    reorderItem = null
                    folderItem = null
                    folderIndex = -1
                    newFolderIcon.active = false
                    reparent(gridview.contentItem)
                    slideMoveAnim.start()
                }
                if (movingItem == launcherItem) {
                    movingItem = null
                }
            }

            x: -2000
            y: -2000
            width: cellWidth
            height: cellHeight
            parent: gridview.contentItem
            scale: newFolderIcon.show && folderIndex == index && itemType == LauncherModel.Application ? 0.5 : (reordering || folderIndex == index ? 1.3 : 1)
            transformOrigin: Item.Center

            onXChanged: if (wrapper.reordering) moved()
            onYChanged: if (wrapper.reordering) moved()

            onClicked: {
                desktop.closeApplicationEnabled = false
                if (dragged) {
                    return
                } else if (itemType == LauncherModel.Folder) {
                    setEditMode(false)
                    showFolder(model.object)
                } else if (launcherEditMode) {
                    setEditMode(false)
                } else if (isUpdating) {
                    // Call launchApplication(), which will send a D-Bus signal
                    // for interested parties (non-store client) to pick it up
                    object.launchApplication()
                } else if (!object.isLaunching) {
                    launcher.launchApplication(object)
                }
            }

            onPressAndHold: setEditMode(true)

            onPressed: {
                // The currentIndex will not be destroyed when we scroll out of view
                currentIndex = index
                if (launcherEditMode) {
                    movingItem = launcherItem
                    previousMovePosition = undefined
                    newIndex = -1
                    newFolderIndex = -1
                    startReorderTimer.restart()
                }
                pressX = mouseX
                pressY = mouseY
                dragged = false
            }

            onPositionChanged: {
                if (!dragged && (Math.abs(pressX-mouseX) > Theme.paddingLarge || Math.abs(pressY-mouseY) > Theme.paddingLarge)) {
                    dragged = true
                }
            }

            onReleased: {
                if (folderIndex >= 0) {
                    if (launcherModel.get(folderIndex).type == LauncherModel.Application) {
                        //% "Folder"
                        var folder = launcherModel.createFolder(folderIndex, qsTrId("lipstick-jolla-home-folder"))
                        if (folder) {
                            folder.iconId = newFolderIcon.iconId

                            var item1Categories = folder.get(0).desktopCategories
                            var item2Categories = model.object.desktopCategories
                            for (var i = 0; i < item1Categories.length; i++) {
                                if (item2Categories.indexOf(item1Categories[i]) >= 0) {
                                    var id = "lipstick-jolla-home-folder_" + item1Categories[i].toLowerCase()
                                    var title = qsTrId(id)
                                    if (title != id) {
                                        folder.title = title
                                        break
                                    }
                                }
                            }
                            launcherModel.moveToFolder(model.object, folder)
                        }
                    } else {
                        launcherModel.moveToFolder(model.object, launcherModel.get(folderIndex))
                    }
                    folderIndex = -1
                    newFolderIcon.active = false
                } else if (!rootFolder && gridview.mapFromItem(launcherItem, 0, 0).y + launcherItem.height/2 > gridview.height) {
                    var parentFolderIndex = launcherModel.parentFolder.indexOf(launcherModel)
                    launcherModel.parentFolder.moveToFolder(model.object, launcherModel.parentFolder, parentFolderIndex+1)
                }

                cancelReordering()
            }

            onCanceled: cancelReordering()

            function reparent(newParent) {
                var newPos = mapToItem(newParent, 0, 0)
                parent = newParent
                x = newPos.x - width/2 * (1-scale)
                y = newPos.y - height/2 * (1-scale)
            }

            function moved() {
                var gridViewPos = gridview.contentItem.mapFromItem(launcherItem, width/2, height/2)
                var item = gridview.itemAt(gridViewPos.x, gridViewPos.y)
                var idx = -1
                var folderIdx = -1
                if (item) {
                    var offset = gridViewPos.x - item.x
                    var folderThreshold = rootFolder && itemType !== LauncherModel.Folder ?
                                gridview.cellWidth / 4 : gridview.cellWidth / 2
                    if (offset < folderThreshold) {
                        if (Math.abs(index - item.modelIndex) > 1 || index > item.modelIndex || item.y !== wrapper.y) {
                            idx = index < item.modelIndex ? item.modelIndex - 1 : item.modelIndex
                            gridview.folderItem = null
                        }
                    } else if (offset >= gridview.cellWidth - folderThreshold) {
                        if (Math.abs(index - item.modelIndex) > 1 || index < item.modelIndex || item.y !== wrapper.y) {
                            idx = index > item.modelIndex ? item.modelIndex + 1 : item.modelIndex
                            gridview.folderItem = null
                        }
                    } else if (item.modelIndex !== index && rootFolder && item.folderItemCount < 99 && itemType !== LauncherModel.Folder) {
                        gridview.folderItem = item
                        folderIdx = item.modelIndex
                    }
                } else if (folderIdx < 0 && idx < 0 && gridViewPos.x >= 0 && gridViewPos.x < gridview.width
                           && gridViewPos.y >= 0) {
                    idx = gridview.count - 1
                }

                if (newIndex !== idx) {
                    newIndex = idx;
                    reorderTimer.restart()
                }
                if (newFolderIndex != folderIdx) {
                    newFolderIndex = folderIdx
                    reorderTimer.restart()
                }
                if (previousMovePosition) {
                    var xdiff = previousMovePosition.x - gridViewPos.x
                    var ydiff = previousMovePosition.y - gridViewPos.y
                    // Have we moved more than paddingSmall/2
                    if (xdiff * xdiff + ydiff * ydiff > Theme.paddingSmall * Theme.paddingSmall / 4) {
                        reorderTimer.restart()
                    }
                }

                if (newFolderIndex != gridview.folderIndex) {
                    gridview.folderIndex = -1
                }

                if (rootFolder) {
                    var modifier = launcher.launcherModifierObject
                    var maxContentY = modifier.contentHeight - desktop.height
                    var globalY = desktop.mapFromItem(launcherItem, 0, 0).y
                    if (globalY <= 0 && modifier.contentY > 0) {
                        modifier.scroll(true)
                    } else if (globalY + height >= dashboard.height && modifier.contentY < maxContentY) {
                        modifier.scroll(false)
                    } else {
                        modifier.stopScrolling()
                    }
                }

                previousMovePosition = gridViewPos
            }

            Behavior on scale {
                NumberAnimation { easing.type: Easing.InOutQuad; duration: 150 }
            }

            ParallelAnimation {
                id: slideMoveAnim
                NumberAnimation { target: launcherItem; property: "x"; to: wrapper.x; duration: 150; easing.type: Easing.InOutQuad }
                NumberAnimation { target: launcherItem; property: "y"; to: wrapper.y; duration: 150; easing.type: Easing.InOutQuad }
            }

            SequentialAnimation {
                id: fadeMoveAnim
                NumberAnimation { target: launcherItem; property: "opacity"; to: 0; duration: 75 }
                ScriptAction { script: { launcherItem.x = wrapper.x; launcherItem.y = wrapper.y } }
                NumberAnimation { target: launcherItem; property: "opacity"; to: 1.0; duration: 75 }
            }

            LauncherIcon {
                id: launcherIcon
                anchors {
                    centerIn: parent
                    verticalCenterOffset: rootFolder && index < gridview.columns ? 0 : Math.round(-launcherText.height/2)
                }
                icon: model.object.iconId
                pressed: down
                opacity: isUpdating && folderItemCount == 0 ? 0.2 : 1.0
                Text {
                    font.pixelSize: Theme.fontSizeExtraLarge
                    font.family: Theme.fontFamilyHeading
                    text: folderItemCount > 0 ? folderItemCount : ""
                    color: Theme.primaryColor
                    anchors.centerIn: parent
                    visible: !isUpdating || model.object.updatingProgress < 0 || model.object.updatingProgress > 100
                    opacity: reorderItem && folderItemCount >= 99 ? 0.3 : 1.0
                }
            }

            Text {
                id: launcherText
                opacity: rootFolder && index < columns && !(launcherEditMode && itemType === LauncherModel.Folder)
                         ? firstRowOpacity : 1.0

                anchors {
                    top: launcherIcon.bottom
                    topMargin: Theme.paddingSmall
                    left: parent.left
                    right: parent.right
                    leftMargin: Theme.paddingSmall/2
                    rightMargin: Theme.paddingSmall/2
                }
                horizontalAlignment: Text.AlignHCenter
                elide: Text.ElideRight

                color: down ? Theme.highlightColor : Theme.primaryColor
                font.pixelSize: Theme.fontSizeTiny
                text: object.title
                visible: !launcherEditMode || itemType === LauncherModel.Folder
            }

            Component {
                id: updatingComponent
                BusyIndicator {
                    anchors.centerIn: launcherIcon
                    running: model.object.isUpdating
                    opacity: running
                    Behavior on opacity { FadeAnimation {} }
                    Label {
                        anchors.centerIn: parent
                        text: model.object.updatingProgress + '%'
                        visible: model.object.isUpdating && model.object.updatingProgress >= 0 && model.object.updatingProgress <= 100
                        font.pixelSize: Theme.fontSizeSmall
                    }
                }
            }

            Component {
                id: uninstallButtonComponent
                UninstallButton {
                    anchors.verticalCenter: launcherIcon.bottom
                    anchors.horizontalCenter: parent.horizontalCenter
                    enabled: launcherEditMode
                    opacity: enabled ? 1.0 : 0.0
                    visible: launcherEditMode && itemType == LauncherModel.Application && appControl.isUninstallable(object.filePath) && !object.isUpdating
                    onClicked: launcher.removeApplication(object.filePath, object.title)
                }
            }
        }
    }
}
